#ifndef VECTOR3D_H
#define VECTOR3D_H

#include <math.h>
#include <string>
#if defined(ENABLE_DISPLAY)
 #include <GL/gl.h>
#endif
#define EPSILON 1e-7
#define RADTODEG 180.0/M_PI
#define DEGTORAD M_PI/180.0

#define DEBUG

class HTMatrix;

float randomNumber(float min, float max);

inline float min(float x, float y)
{  return ( (x<y) ? x : y );  }

inline float max(float x, float y)
{  return ( (x>y) ? x : y );  }

inline int sign(float x)
{  return ( (x>0) ? 1 : -1 );  }

class Vector2D
{
  public:
   float x, y;

   Vector2D();
   Vector2D(const Vector2D &v);
   Vector2D(float x0, float y0);
   void set(float x0, float y0);
   Vector2D operator + () const;
   Vector2D operator - () const;
   friend Vector2D operator + (const Vector2D &v1, const Vector2D &v2); 
   friend Vector2D operator - (const Vector2D &v1, const Vector2D &v2); 
   void operator+= (const Vector2D& v);
   void operator-= (const Vector2D& v);
   void operator*= (float s);
   void operator/= (float s);
   Vector2D operator * (float a);
   friend Vector2D operator * (float a , Vector2D v);
   Vector2D operator / (float a);
   float operator () (unsigned int i) const;
   float& operator () (unsigned int i);
   float operator [] (unsigned int i) const;
   float& operator [] (unsigned int i);

   bool operator==(const Vector2D& v) const;
   bool operator!=(const Vector2D& v)  const { return !(*this == v); }

   Vector2D orthogonalVector();

   float norm();
   int normalize();
   void print(const char * name);
};


float dotProduct(const Vector2D &v1, const Vector2D &v2);
float norm(const Vector2D &v);

class Vector3D
{
  public:
   float x, y, z;

   Vector3D();
   Vector3D(const Vector3D &v);
   Vector3D(float x0, float y0, float z0);
   void set(float x0, float y0, float z0);
   Vector3D operator + () const;
   Vector3D operator - () const;
   friend Vector3D operator + (const Vector3D &v1, const Vector3D &v2); 
   friend Vector3D operator - (const Vector3D &v1, const Vector3D &v2); 
   void operator += (const Vector3D& v);
   void operator -= (const Vector3D& v);
   void operator *= (float s);
   void operator /= (float s);
   Vector3D operator * (float a);
   friend Vector3D operator * (float a , Vector3D v); 
   friend Vector3D operator * (const HTMatrix &M, Vector3D v); 
   Vector3D operator / (float a);
   float operator () (unsigned int i) const;
   float& operator () (unsigned int i);
   float operator [] (unsigned int i) const;
   float& operator [] (unsigned int i);
   bool operator==(const Vector3D& v) const;
   bool operator!=(const Vector3D& v)  const { return !(*this == v); }

   Vector3D orthogonalVector();

   float norm();
   int normalize();
   void print(const char * name);
};

float dotProduct(const Vector3D &v1, const Vector3D &v2);
float norm(const Vector3D &v);
Vector3D crossProduct(const Vector3D &v1, const Vector3D &v2);
float pointToTriangleDistance(const Vector3D &point, const Vector3D &p0, const Vector3D &p1, const Vector3D &p2, Vector3D &closestPoint);

//! Homogeneous transform matrix class:
class HTMatrix
{
  public: 
    float m11, m12, m13, m14;
    float m21, m22, m23, m24;
    float m31, m32, m33, m34;

    HTMatrix()
    {
      m11= m22= m33= 1.0;
      m12= m13= m14= m21= m23= m24 = m31= m32= m34= 0.0;
    }

    HTMatrix operator * (const HTMatrix &mat);
    void operator *= (const HTMatrix &mat);
    float operator () (unsigned int i, unsigned int j) const;
    float& operator () (unsigned int i, unsigned int j);
    Vector3D operator * (const Vector3D &v);

    void setIdentity();
    void setTranslation(float x, float y, float z);
    void setRotationFromAngleAndAxis(float theta, float x, float y, float z);
    void setRotationFromYPRangles(float yaw, float pitch, float roll);
    int setFromPointAndAxis(const Vector3D &c, float x, float y, float z);
    int setFromDHparameters(float a, float alpha, float d, float theta);
    void invert();
    Vector3D firstColumn();
    Vector3D secondColumn();
    Vector3D thirdColumn();
    Vector3D fourthColumn();
    void setFirstColumn(Vector3D &v);
    void setSecondColumn(Vector3D &v);
    void setThirdColumn(Vector3D &v);
    void setFourthColumn(Vector3D &v);
    void setFrom4x4Array(float mat[4][4]);
#if defined(ENABLE_DISPLAY)
    void convertToOpenGLFormat(GLfloat mat[16]);
#endif
    void print(std::string name);
};

Vector3D rotationMult(const HTMatrix &mat, const Vector3D &v);
HTMatrix inverse(const HTMatrix &mat);


//! Arbitrary size matrix class:
class Matrix
{
  private:
    float *data;
    unsigned int nbRows_, nbColumns_;

  public:
    unsigned int nbRows() const { return nbRows_; }
    unsigned int nbColumns() const { return nbColumns_; }
    Matrix();
    Matrix(unsigned int nbRows, unsigned int nbColumns);
    Matrix(const Matrix &matrix);
    int resize(unsigned int nbRows, unsigned int nbColumns);
    Matrix & operator = (const Matrix &mat);
    Matrix operator * (const Matrix &mat);
    void operator *= (const Matrix &mat);
    float operator () (unsigned int i, unsigned int j) const;
    float& operator () (unsigned int i, unsigned int j);
    int setToZero();
    int setToIdentity(unsigned int size);
    void print(std::string name);
    int setToRandom(float min, float max);
};

int inverse(Matrix &mat, Matrix &result);

#endif
