#if defined(ENABLE_DISPLAY)
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glu.h>
#include <camera.h>


Camera::Camera()
{
  mode= LOCKED_CAMERA;

  fov= 45.0;
  focusPoint.set( 0.0, 0.0, 0.9 );
  radius= 3.0;
  elevation= M_PI/12.0;
  azimuth= 0.0;

  updateCartesianCoordinates();
}

void Camera::updateCartesianCoordinates()
{
  position.x= focusPoint.x + radius*cos(elevation)*cos(azimuth);
  position.y= focusPoint.y + radius*cos(elevation)*sin(azimuth);
  position.z= focusPoint.z + radius*sin(elevation);

  upVector.x= -sin(elevation)*cos(azimuth);
  upVector.y= -sin(elevation)*sin(azimuth);
  upVector.z= cos(elevation);

  viewDir= focusPoint - position;
  viewDir.normalize();

  rightVector= crossProduct(viewDir, upVector);
  rightVector.normalize();
}


//! Switches between free motion and locked mode.
void Camera::switchMode()
{ 
  switch(mode)
  {
    case FREE_CAMERA:
      mode= LOCKED_CAMERA;
      focusPoint.set( 0.0, 0.0, 0.0 );
      radius= norm( position - focusPoint);
      azimuth= atan2( position.y, position.x );
      elevation= acos(position.z / radius);
    break;
    case LOCKED_CAMERA:
      mode= FREE_CAMERA; 
      updateCartesianCoordinates();
    break;
  }
}




//! Performs the OpenGL MODELVIEW transformation corresponding to the camera's current position.
void Camera::render(void)
{
  switch(mode)
  {
    case FREE_CAMERA:
      gluLookAt(position.x,  position.y,  position.z,
                position.x + viewDir.x, position.y + viewDir.y, position.z + viewDir.z,
                upVector.x,  upVector.y,  upVector.z); 
    break;
    case LOCKED_CAMERA:
      updateCartesianCoordinates();
      gluLookAt(position.x,  position.y,  position.z,
                focusPoint.x, focusPoint.y, focusPoint.z,
                upVector.x,  upVector.y,  upVector.z); 
    break;
  }
}

//! Adds the given coordinates to the camera position.
void Camera::move(float x, float y, float z)
{
  position.x+= x;
  position.y+= y;
  position.z+= z;

  focusPoint.x+= x;
  focusPoint.y+= y;
  focusPoint.z+= z;
}


//! Rotates the camera around an axis parallel to the world's vertical axis, passing through the camera's center.
//! \param angle rotation angle in degrees.
void Camera::pan(float angle)
{
  float ax, ay;

  angle*= DEGTORAD;

  ax= rightVector.x;
  ay= rightVector.y;
  rightVector.x=  cos(angle)*ax - sin(angle)*ay;
  rightVector.y=  sin(angle)*ax + cos(angle)*ay;
  rightVector.z= 0;

  rightVector.normalize();

  ax= viewDir.x;
  ay= viewDir.y;

  viewDir.x= cos(angle)*ax - sin(angle)*ay;
  viewDir.y= sin(angle)*ax + cos(angle)*ay;
  viewDir.normalize();

  upVector= crossProduct(rightVector, viewDir);
  upVector.normalize();
}


//! Rotates the camera around the rightVector vector.
//! \param angle rotation angle in degrees.
void Camera::tilt(float angle)
{
  float c, s;

  c= cos(angle*M_PI/180);
  s= sin(angle*M_PI/180);

  upVector= c*upVector + s*viewDir;
  upVector.normalize();

  viewDir= crossProduct(upVector, rightVector);
  viewDir.normalize();
}


//! Moves the camera in the direction of its look.
//! The motion along z axis is removed.
//! \param distance the length of the motion along the camera look direction (can be either positive or negative)
void Camera::moveForward(float distance)
{
  Vector3D v(viewDir.x, viewDir.y, 0);

  v.normalize();
  position+= distance*v;
  focusPoint+= distance*v;
}


void Camera::moveSideways(float distance)
{
  position+= distance*rightVector;

  focusPoint+= distance*rightVector;
}


void Camera::zoom(float distance)
{
  radius+= distance;

  if(radius < 0.2)
  {  radius= 0.2;  }
}

//! Changes the azimuth angle of the camera.
//! \param angle rotation angle in degrees
void Camera::changeAzimuth(float angle)
{
  angle*= DEGTORAD;

  azimuth+= angle;

  if(azimuth > 2*M_PI)
  {
    azimuth= azimuth - 2*M_PI;
  }

  if(azimuth < 0)
  {
    azimuth= 2*M_PI + azimuth;
  }
}

//! Changes the elevation angle of the camera.
//! \param angle rotation angle in degrees
void Camera::changeElevation(float angle)
{
  angle*= DEGTORAD;
  elevation+= angle;

  if(elevation > 85*DEGTORAD)
  {  elevation= 85*DEGTORAD;  }

  if(elevation < -85*DEGTORAD)
  {  elevation= -85*DEGTORAD;  }
}


float Camera::distanceFromFocusPoint()
{
  return norm(focusPoint - position);
}

#endif
