#if defined(ENABLE_DISPLAY)
#include <GL/gl.h>
#include <GL/freeglut.h>
#endif
#include <file_reader.h>
#include <graphics.h>
#include <robot.h>

//! Converts a DOFtopology to a std::string.
std::string DOFtopologyToString(DOFtopology topology)
{
  switch(topology)
  {
    case R_TOPOLOGY:
      return "R_TOPOLOGY";
    break;
    case S_TOPOLOGY:
      return "S_TOPOLOGY";
    break;
    case SO3_TOPOLOGY:
      return "SO3_TOPOLOGY";
    break;
  }

  return "UNDEFINED";
}


//! Converts a DOFstate to a std::string.
std::string DOFstateToString(DOFstate state)
{
  switch(state)
  {
    case INDEPENDENT:
      return "INDEPENDENT";
    break;
    case CONSTRAINED:
      return "CONSTRAINED";
    break;
    case BLOCKED:
      return "BLOCKED";
    break;
  }

  return "UNDEFINED";
}



//! Converts a DOFunit to a std::string.
std::string DOFunitToString(DOFunit unit)
{
  switch(unit)
  {
    case METER:
      return "METER";
    break;
    case RADIAN:
      return "RADIAN";
    break;
    case DEGREE:
      return "DEGREE";
    break;
  }

  return "UNDEFINED";
}


//! Default constructor of the class DOF.
DOF::DOF()
{
  state= INDEPENDENT;
  _unit = RADIAN;
  topology = R_TOPOLOGY;
  _value= 0.0;
  _velocity = 0.0;
  _acceleration = 0.0;
  _min= -1.0;
  _max=  1.0;
}

//! Sets the value of the DOF.
//! \param value the value to set
//! \return 1 in case of success, 0 otherwise
int DOF::setValue(float value)
{
  if(topology==R_TOPOLOGY || topology==S_TOPOLOGY)
  {
    if( (value < _min) || (value > _max) )
    {
      printf("%s: %d: DOF::setValue(): input value (%f) exceeds DOF bounds (%f, %f).\n", __FILE__, __LINE__,value,_min,_max);
      return 0;
    }
  }

  _value= value;

  return 1;
}

//! Sets the velocity of the DOF.
//! \param velocity the velocity to set
//! \return 1 in case of success, 0 otherwise
int DOF::setVelocity(float velocity)
{
  _velocity= velocity;
  return 1;
}

//! Sets the acceleration of the DOF.
//! \param acceleration the acceleration to set
//! \return 1 in case of success, 0 otherwise
int DOF::setAcceleration(float acceleration)
{
  _acceleration= acceleration;
  return 1;
}

//! Sets the unit of the DOF.
//! \param unit the unit to set
//! \return 1 in case of success, 0 otherwise
int DOF::setUnit(DOFunit unit)
{
  _unit= unit;

  return 1;
}

//! Gets the DOF minimal bound.
float DOF::min()
{
  return _min;
}

//! Gets the DOF maximal bound.
float DOF::max()
{
  return _max;
}

//! Gets the DOF current value.
float DOF::value()
{
  return _value;
}

//! Gets the DOF current value.
float DOF::velocity()
{
  return _velocity;
}

//! Gets the DOF current value.
float DOF::acceleration()
{
  return _acceleration;
}

//! Gets the DOF current unit.
DOFunit DOF::unit()
{
  return _unit;
}

//! Sets the DOF bounds.
//! \param min minimal bound
//! \param max maximal bound
//! \return 1 in case of success, 0 otherwise
int DOF::setBounds(float min, float max)
{
  _min= min;
  _max= max;

  return 1;
}

//! Gets the DOF bounds.
//! \param min where to copy the minimal bound
//! \param max where to copy the  maximal bound
//! \return 1 in case of success, 0 otherwise
int DOF::getBounds(float &min, float &max)
{
  min= _min;
  max= _max;

  return 1;
}

//! Converts a JointType to a std::string.
std::string jointTypeToString(JointType type)
{
  switch(type)
  {
    case BASE:
      return "BASE";
    break;
    case FREEFLYER:
      return "FREEFLYER";
    break;
    case REVOLUTE:
      return "REVOLUTE";
    break;
    case S_REVOLUTE:
      return "S_REVOLUTE";
    break;
    case PRISMATIC:
      return "PRISMATIC";
    break;
  }

  return "UNDEFINED";
}

Joint::Joint(JointType _type)
{
  _name= "none";
  previous= NULL;
  up_to_date= false;
  selected= false;
  type= _type;
  switch(type)
  {
    case BASE:
      q.resize(3);
      q[0].setUnit(METER);
      q[0].topology= R_TOPOLOGY;
      q[1].setUnit(METER);
      q[1].topology= R_TOPOLOGY;
      q[2].setUnit(RADIAN);
      q[2].topology= S_TOPOLOGY;
    break;
    case FREEFLYER:
      q.resize(6);
      q[0].setUnit(METER);
      q[0].topology= R_TOPOLOGY;
      q[1].setUnit(METER);
      q[1].topology= R_TOPOLOGY;
      q[2].setUnit(METER);
      q[2].topology= R_TOPOLOGY;
      q[3].setUnit(RADIAN);
      q[3].topology= SO3_TOPOLOGY;
      q[4].setUnit(RADIAN);
      q[4].topology= SO3_TOPOLOGY;
      q[5].setUnit(RADIAN);
      q[5].topology= SO3_TOPOLOGY;
    break;
    case REVOLUTE: 
      q.resize(1);
      q[0].setUnit(RADIAN);
      q[0].topology= R_TOPOLOGY;
    break;   
    case S_REVOLUTE: 
      q.resize(1);
      q[0].setUnit(RADIAN);
      q[0].topology= S_TOPOLOGY;
    break;     
    case PRISMATIC: 
      q.resize(1);
      q[0].setUnit(METER);
      q[0].topology= R_TOPOLOGY;
    break;     
    default:
      printf("%s: %d: Joint::Joint(): invalid joint type.\n", __FILE__, __LINE__);
    break;   
  }
}


Joint::~Joint()
{
  _name.clear();
}

//! Changes the joint's name.
void Joint::rename(const std::string name)
{
  _name= name;
}


//! Prints some information about the joint.
//! \return 1 in case of success, 0 otherwise
int Joint::printInfo()
{
   unsigned int i, size, minSize, maxSize, valueSize, unitSize;
   std::string str, space;

   std::cout << "joint: " << _name << std::endl;

   std::cout << "\t" << jointTypeToString(type) << std::endl;

   if(previous!=NULL)
   {  std::cout << "\tprevious: " << previous->name() << std::endl;   }
   else
   {  std::cout << "\tprevious: none" << std::endl;   }

   if(!next.empty())
   {
      std::cout << "\tnext: " << std::endl;
      for(i=0; i<next.size(); ++i)
      {
        std::cout << "\t " << next[i]->name() << std::endl;
      }
   }
   else
   {
      std::cout << "\tnext: none" << std::endl;
   }

   minSize= maxSize= 3;
   valueSize= 5;
   unitSize= 5;

   //get the maximum size of each parameter to give then the same size to each string:
   for(i=0; i<q.size(); ++i)
   {
     str= convertToString(q[i].min());
     size= str.size();
     if(size > minSize)
     {  minSize= size; }

     str= convertToString(q[i].max());
     size= str.size();
     if(size > maxSize)
     {  maxSize= size; }

     str= convertToString(q[i].value());
     size= str.size();
     if(size > valueSize)
     {  valueSize= size; }

     str= DOFunitToString(q[i].unit());
     size= str.size();
     if(size > unitSize)
     {  unitSize= size; }
   }

   space= "   ";

   str= "min"; 
   if(str.size() < minSize)
   {   str.resize(minSize,' ');  }
   std::cout << "\tq= [  " << str << space;

   str= "max";
   if(maxSize > str.size())
   {  str.resize(maxSize,' ');  }
   std::cout << str << space;

   str= "value";
   if(valueSize > str.size())
   {  str.resize(valueSize,' ');  }
   std::cout << str << space;

   str= "unit";
   if(unitSize > str.size())
   {  str.resize(unitSize,' ');  }

   std::cout << str << space << "state  ]" << std::endl;

   for(i=0; i<q.size(); ++i)
   {
     str= convertToString(q[i].min());
     str.resize(minSize,' ');
     std::cout << "\t      " << str << space;

     str= convertToString(q[i].max());
     str.resize(maxSize,' ');
     std::cout << str << space;

     str= convertToString(q[i].value());
     str.resize(valueSize,' ');
     std::cout << str << space;

     str= DOFunitToString(q[i].unit());
     str.resize(unitSize,' ');
     std::cout << str << space;

     std::cout << DOFstateToString(q[i].state) << std::endl;
   }

   return 1;
}

#if defined(ENABLE_DISPLAY)
//! Draws the joint frame.
//! \param length length of the axes
//! \return 1 in case of success, 0 otherwise
int Joint::draw(float length)
{
   GLboolean enable_lighting;
   GLint line_width;

   if(selected)
   {  drawFrame(frame, 1.1*length);  }
   else
   {  drawFrame(frame, length);  }
   return 1;

   glGetBooleanv(GL_LIGHTING, &enable_lighting);
   glGetIntegerv(GL_LINE_WIDTH, &line_width);

   if(selected)
   {  glLineWidth(6);  }
   else
   {  glLineWidth(2);  }

   Vector3D origin, xAxis, yAxis, zAxis;

   xAxis = frame.firstColumn();
   yAxis = frame.secondColumn();
   zAxis = frame.thirdColumn();
   origin= frame.fourthColumn();


   glDisable(GL_LIGHTING);
   glColor3f(1,0,0);
   glBegin(GL_LINES);
     glVertex3d(origin.x, origin.y, origin.z);
     glVertex3d(origin.x + length*xAxis.x, origin.y + length*xAxis.y, origin.z + length*xAxis.z );
   glEnd();


   glColor3f(0,1,0);
   glBegin(GL_LINES);
     glVertex3d(origin.x, origin.y, origin.z);
     glVertex3d(origin.x + length*yAxis.x, origin.y + length*yAxis.y, origin.z + length*yAxis.z );
   glEnd();

   glColor3f(0,0,1);
   glBegin(GL_LINES);
     glVertex3d(origin.x, origin.y, origin.z);
     glVertex3d(origin.x + length*zAxis.x, origin.y + length*zAxis.y, origin.z + length*zAxis.z );
   glEnd();

   glLineWidth(2);


   if(enable_lighting)
   {  glEnable(GL_LIGHTING);  }
   else
   {  glDisable(GL_LIGHTING);  }

   glLineWidth(line_width);

   return 1;
}
#endif
//! Computes the transform matrix due to the DOFs of the joint.
//! \return 1 in case of success, 0 otherwise
int Joint::computeTransform()
{
  HTMatrix tmp;
  T.setIdentity();
  switch(type)
  {
    case BASE: 
      tmp.setTranslation(q[0].value(), q[1].value(), 0);
      T.setRotationFromYPRangles(q[2].value(), 0, 0);
      T= tmp*T;
    break;
    case FREEFLYER: 
      tmp.setTranslation(q[0].value(), q[1].value(), q[2].value());
      T.setRotationFromYPRangles(q[5].value(), q[4].value(), q[3].value());
      T= tmp*T;
    break;
    case REVOLUTE: case S_REVOLUTE:
      T.setRotationFromAngleAndAxis(q[0].value(), 0, 0, 1);
    break;
    case PRISMATIC:
      T.setTranslation(0, 0, q[0].value());
    break;
  }

  return 1;
}

//! \return the number of active DOFs of the joint
unsigned int Joint::nbDOFs()
{
   unsigned int i, count;

   count= 0;

   for(i=0; i<q.size(); ++i)
   {
     if(q[i].state==BLOCKED)
     {  continue;  }
     count++;
   }


   return count;
}

//! Sets the value of one of the joint DOFs (passive or active).
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED state)
//! \param x value to set
//! \return 1 in case of success, 0 otherwise
int Joint::setDOFvalue(unsigned int index, float x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(count==index)
    { 
       if(!q[i].setValue(x))
       {
         printf("%s: %d: Joint::setDOFvalue(): input DOF value exceeds one of the DOF bounds (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
         return 0;
       }
       return 1;
    }
    count++;
  }
  printf("Joint::setDOFvalue(): %d\n", index);
  printf("%s: %d: Joint::setDOFvalue(): input DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());

  return 0;
}

//! Gets the value of one of the joint DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x value of the DOF
//! \return 1 in case of success, 0 otherwise
int Joint::getDOFvalue(unsigned int index, float &x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(q[i].state==CONSTRAINED)
    {  continue;  }

    if(count==index)
    { 
       x= q[i].value();
       return 1;
    }
    count++;
  }

  printf("%s: %d: Joint::getDOFvalue(): DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
  return 0;
}


//! Sets the velocity of one of the joint DOFs (passive or active).
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED state)
//! \param x velocity to set
//! \return 1 in case of success, 0 otherwise
int Joint::setDOFvelocity(unsigned int index, float x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(count==index)
    { 
       if(!q[i].setVelocity(x))
       {
         printf("%s: %d: Joint::setDOFvelocity(): input DOF value exceeds one of the DOF bounds (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
         return 0;
       }
       return 1;
    }
    count++;
  }
  printf("Joint::setDOFvelocity(): %d\n", index);
  printf("%s: %d: Joint::setDOFvelocity(): input DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());

  return 0;
}


//! Gets the velocity of one of the joint DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x velocity of the DOF
//! \return 1 in case of success, 0 otherwise
int Joint::getDOFvelocity(unsigned int index, float &x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(q[i].state==CONSTRAINED)
    {  continue;  }

    if(count==index)
    { 
       x= q[i].velocity();
       return 1;
    }
    count++;
  }

  printf("%s: %d: Joint::getDOFvelocity(): DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
  return 0;
}

//! Sets the acceleration  of one of the joint DOFs (passive or active).
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED state)
//! \param x acceleration to set
//! \return 1 in case of success, 0 otherwise
int Joint::setDOFacceleration(unsigned int index, float x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(count==index)
    { 
       if(!q[i].setAcceleration(x))
       {
         printf("%s: %d: Joint::setDOFacceleration(): input DOF value exceeds one of the DOF bounds (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
         return 0;
       }
       return 1;
    }
    count++;
  }
  printf("Joint::setDOFacceleration(): %d\n", index);
  printf("%s: %d: Joint::setDOFacceleration(): input DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());

  return 0;
}


//! Gets the accelaration of one of the joint DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x acceleration of the DOF
//! \return 1 in case of success, 0 otherwise
int Joint::getDOFacceleration(unsigned int index, float &x)
{
  unsigned int i, count;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(q[i].state==CONSTRAINED)
    {  continue;  }

    if(count==index)
    { 
       x= q[i].acceleration();
       return 1;
    }
    count++;
  }

  printf("%s: %d: Joint::getDOFacceleration(): DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());
  return 0;
}


//! Increments the value of one of the joint DOFs (passive or active).
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED state)
//! \param ratio increment ratio (the increment will be ratio*(qmax-qmin))
//! \return 1 in case of success, 0 otherwise
int Joint::incrementDOFvalue(unsigned int index, float ratio)
{
  unsigned int i, count;
  float x0, xmin, xmax, dx;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(count==index)
    { 
       x0= q[i].value();
       q[i].getBounds(xmin, xmax);
       dx= ratio*(xmax - xmin);

       if( (x0 + dx) > xmax )
       {
         if(q[i].topology==R_TOPOLOGY)
         {   q[i].setValue(xmax);    }
         else
         {   q[i].setValue(xmin + (dx -(xmax-x0)) );   } 
       }
       else
       {  q[i].setValue(x0 + dx);    }

       return 1;
    }
    count++;
  }
  printf("Joint::incrementDOFvalue(): %d\n", index);
  printf("%s: %d: Joint::incrementDOFvalue(): DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());

  return 0;
}


//! Decrements the value of one of the joint DOFs (passive or active).
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED state)
//! \param ratio decrement ratio (the decrement will be ratio*(qmax-qmin))
//! \return 1 in case of success, 0 otherwise
int Joint::decrementDOFvalue(unsigned int index, float ratio)
{
  unsigned int i, count;
  float x0, xmin, xmax, dx;

  count= 0;
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==BLOCKED)
    {  continue;  }

    if(count==index)
    { 
       x0= q[i].value();
       q[i].getBounds(xmin, xmax);
       dx= ratio*(xmax - xmin);

       if( (x0 - dx) < xmin )
       {
         if(q[i].topology==R_TOPOLOGY)
         {   q[i].setValue(xmin);     }
         else
         {   q[i].setValue(xmax - (dx -(x0-xmin)) );     } 
       }
       else
       {  q[i].setValue(x0 - dx);    }

       return 1;
    }
    count++;
  }
  printf("Joint::decrementDOFvalue(): %d\n", index);
  printf("%s: %d: Joint::decrementDOFvalue(): DOF index exceeds the joint number of DOFs (joint \"%s\").\n", __FILE__, __LINE__,_name.c_str());

  return 0;
}

