#include <iostream>
#include <sstream>
//#include <libxml/xmlreader.h>
#if defined(ENABLE_DISPLAY)
#include <GL/gl.h>
#include <GL/freeglut.h>
#endif
#include <robot.h>
#include <vector3D.h>
#include <file_reader.h>
#include <graphics.h>


Robot::Robot()
{
  _file= "none";
  _name= "none";
  models_directory= "./";
}

Robot::~Robot()
{
  unsigned int i;
  _name.clear();
 
  for(i=0; i<bodies.size(); ++i)
  {
    delete bodies[i];
  }
  for(i=0; i<joints.size(); ++i)
  {
    delete joints[i];
  }

  bodies.clear();
  joints.clear();
}

//! Changes the robot's name.
void Robot::rename(const std::string name)
{
  _name= name;
}

//! Gets the robot's description file (XML).
std::string Robot::file()
{  return _file;  }

//! Gets the robot's description path (XML).
std::string Robot::path()
{  return _path;  }

//! Gets the robot's name.
std::string Robot::name()
{  return _name;  }

//! Gets the robot's number of bodies.
unsigned int Robot::nbBodies()
{  return bodies.size();  }

//! Gets the robot's number of joints.
unsigned int Robot::nbJoints()
{  return joints.size();  }

//! Draws the robot.
//! \param smooth use smooth (true) or flat (false) shading
//! \return 1 in case of success, 0 otherwise
int Robot::draw(bool smooth)
{
  unsigned int i;
 
  for(i=0; i<bodies.size(); ++i)
  {
    bodies[i]->draw(smooth);
  }

  return 1;
}

//! Draws the robot with black, unlit bodies.
//! This function is meant to be used to display planar shadows.
//! \return 1 in case of success, 0 otherwise
int Robot::drawInBlack()
{
  unsigned int i;
 
  for(i=0; i<bodies.size(); ++i)
  {
    bodies[i]->drawInBlack();
  }

  return 1;
}

//! Prints some information about the robot's components.
void Robot::printInfo()
{
  unsigned int i;

  std::cout << "Robot: " << _name << std::endl;
  std::cout << "  file=  " << _file << std::endl;
  std::cout << "  nb_bodies= " << bodies.size() << std::endl;
  std::cout << "  nb_joints= " << joints.size() << std::endl;

//   cout << " bodies: " << std::endl;
//   for(i=0; i<bodies.size(); ++i)
//   {
//     cout << "\t   " << bodies[i]->name() << endl;
//     if(bodies[i]->joint()!=NULL)
//     {
//       cout << "\t     joint:" << bodies[i]->joint()->name() << endl;
//     }
//     else
//     {
//       cout << "\t     joint: none" << endl;
//     }
//   }

  for(i=0; i<joints.size(); ++i)
  {
    joints[i]->printInfo();
  }

  for(i=0; i<constraints.size(); ++i)
  {
    std::cout << "constraint:" << std::endl;
    std::cout << "\tmaster" << constraints[i].master->name() << std::endl;
    std::cout << "\tslave" << constraints[i].slave->name() << std::endl;
    std::cout << "\tmaster_dof " << constraints[i].master_dof << std::endl;
    std::cout << "\tslave_dof " << constraints[i].slave_dof << std::endl;
  }

}

//! Draws the robot joints as small frames.
//! \param length length of the axes
//! \return 1 in case of success, 0 otherwise
int Robot::drawJoints(float length)
{
  unsigned int i;
  Vector3D p1, p2;

#ifdef ENABLE_DISPLAY
  for(i=0; i<joints.size(); ++i)
  {
    joints[i]->draw(length);
  }
#endif
/*
  for(i=1; i<joints.size(); ++i)
  {
    p1= joints[i-1]->frame.fourthColumn();
    p2= joints[i]->frame.fourthColumn();
    drawCylinder(p1, p2, 0.1);
  }
*/
  return 1;
}



//! Sets the current DOF value vector of the robot.
//! The input vector must only contain the values of the DOFS in INDEPENDENT state.
//! \param q a std::vector containing the DOF values of the robot (INDEPENDENT state) in the desired configuration
//! \return 1 in case of success, 0 otherwise
int Robot::setConfiguration(std::vector<float> &q)
{
   unsigned int i, j, count;

   if(q.size()!=nbDOFs())
   {
     printf("%s: %d: Robot::setConfiguration(): input vector q does not have the good size (%ud instead of %ud).\n", __FILE__, __LINE__,q.size(), nbDOFs());
     return 0;
   }

   count= 0;
   for(i=0; i<joints.size(); ++i)
   { 
      for(j=0; j<joints[i]->q.size(); j++)
      {
         if(joints[i]->q[j].state==BLOCKED || joints[i]->q[j].state==CONSTRAINED)
         {  continue;  }

         joints[i]->q[j].setValue(q[count]);
         count++;
      }
   }

   updateConfiguration();

   return 1;
}


//! Gets the current DOF value vector of the robot.
//! DOFs in BLOCKED state are not taken into account.
//! \param q a std::vector variable the robot's DOF values will be copied in
//! \return 1 in case of success, 0 otherwise
int Robot::getConfiguration(std::vector<float> &q)
{
   unsigned int i, j, count;

   q.resize(nbDOFs());

   count= 0;
   for(i=0; i<joints.size(); ++i)
   { 
      for(j=0; j<joints[i]->q.size(); j++)
      {
         if(joints[i]->q[j].state==BLOCKED || joints[i]->q[j].state==CONSTRAINED)
         {  continue;  }

         q[count]= joints[i]->q[j].value();
	 count++;
      }
   }

   return 1;
}

//! Gets the current DOF value vector of the robot.
//! DOFs in BLOCKED state are not taken into account.
//! \param q a std::vector variable the robot's DOF values will be copied in
//! \return 1 in case of success, 0 otherwise
int Robot::getVelocity(std::vector<float> &q)
{
   unsigned int i, j, count;

   q.resize(nbDOFs());

   count= 0;
   for(i=0; i<joints.size(); ++i)
   { 
      for(j=0; j<joints[i]->q.size(); j++)
      {
         if(joints[i]->q[j].state==BLOCKED || joints[i]->q[j].state==CONSTRAINED)
         {  continue;  }

         q[count]= joints[i]->q[j].velocity();
	 count++;
      }
   }

   return 1;
}

//! Gets the current DOF value vector of the robot.
//! DOFs in BLOCKED state are not taken into account.
//! \param q a std::vector variable the robot's DOF values will be copied in
//! \return 1 in case of success, 0 otherwise
int Robot::getAcceleration(std::vector<float> &q)
{
   unsigned int i, j, count;

   q.resize(nbDOFs());

   count= 0;
   for(i=0; i<joints.size(); ++i)
   { 
      for(j=0; j<joints[i]->q.size(); j++)
      {
         if(joints[i]->q[j].state==BLOCKED || joints[i]->q[j].state==CONSTRAINED)
         {  continue;  }

         q[count]= joints[i]->q[j].acceleration();
	 count++;
      }
   }

   return 1;
}


//! Selects one of the robot DOFs.
//! \param index index of the DOF to select (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \return 1 in case of success, 0 otherwise
int Robot::selectDOF(unsigned int index)
{
   unsigned int i, j, count;

   for(i=0; i<joints.size(); ++i)
   {
     joints[i]->selected= false;
   }


   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
          joints[i]->selected= true;
          return 1;
        }
        count++;
      }
   }

   printf("%s: %d: Robot::selectDOF(): DOF index (%d) exceeds the robot number of DOFs.\n", __FILE__, __LINE__,index);
   return 0;
}


//! Gets the vector of all the robot's active DOFs.
//! \param q a std::vector variable the robot's DOF vector will be copied in
//! \return 1 in case of success, 0 otherwise
int Robot::getDOFs(std::vector<DOF> &q)
{
  unsigned int i, j, count;

  q.resize(nbDOFs());

  count= 0;
  for(i=0; i<joints.size(); ++i)
  {
     for(j=0; j<joints[i]->q.size(); j++)
     {
       if(joints[i]->q[j].state==BLOCKED)
       {  continue;  }

       if(joints[i]->q[j].state==CONSTRAINED)
       {  continue;  }

       if(count >= q.size())
       {
         return 0;
       }

       q[count]= joints[i]->q[j];
       count++;
    }
  }

  return 1;
}

//! Gets the bounds on one of the robot DOF value.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param min minimal bound on the DOF value
//! \param max maximal bound on the DOF value
//! \return 1 in case of success, 0 otherwise
int Robot::getDOFbounds(unsigned int index, float &min, float &max)
{
   unsigned int i, j, count;

   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
          min= joints[i]->q[j].min();
          max= joints[i]->q[j].max();
          return 1;
        }
        count++;
      }
   }

   printf("%s: %d: Robot::getDOFbounds(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}



//! Sets the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x value to set
//! \return 1 in case of success, 0 otherwise
int Robot::setDOFvalue(unsigned int index, float x)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
//           if( (x<joints[i]->q[j].min()) || (x>joints[i]->q[j].max()) )
//           {
//             printf("%s: %d: Robot::setDOF(): DOF value exceeds one of the bound values.\n", __FILE__, __LINE__);
//             printf("x= %f  qmin= %f qmax= %f\n", x, joints[i]->q[j].min(), joints[i]->q[j].max());  
//             return 0;             
//           }
//           joints[i]->q[j].setValue(x);
          joints[i]->setDOFvalue(j, x);

          updatePassiveJoints();
          updateConfiguration();
          return 1;
        }

        count++;
      }
   }

   printf("%s: %d: Robot::setDOFvalue(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}


//! Sets the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x value to set
//! \return 1 in case of success, 0 otherwise
int Robot::setDOFvelocity(unsigned int index, float x)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
//           if( (x<joints[i]->q[j].min()) || (x>joints[i]->q[j].max()) )
//           {
//             printf("%s: %d: Robot::setDOF(): DOF value exceeds one of the bound values.\n", __FILE__, __LINE__);
//             printf("x= %f  qmin= %f qmax= %f\n", x, joints[i]->q[j].min(), joints[i]->q[j].max());  
//             return 0;             
//           }
//           joints[i]->q[j].setValue(x);
          joints[i]->setDOFvelocity(j, x);

          updatePassiveJoints();
          updateConfiguration();
          return 1;
        }

        count++;
      }
   }

   printf("%s: %d: Robot::setDOFvelocity(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}

//! Sets the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x value to set
//! \return 1 in case of success, 0 otherwise
int Robot::setDOFacceleration(unsigned int index, float x)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
//           if( (x<joints[i]->q[j].min()) || (x>joints[i]->q[j].max()) )
//           {
//             printf("%s: %d: Robot::setDOF(): DOF value exceeds one of the bound values.\n", __FILE__, __LINE__);
//             printf("x= %f  qmin= %f qmax= %f\n", x, joints[i]->q[j].min(), joints[i]->q[j].max());  
//             return 0;             
//           }
//           joints[i]->q[j].setValue(x);
          joints[i]->setDOFacceleration(j, x);

          updatePassiveJoints();
          updateConfiguration();
          return 1;
        }

        count++;
      }
   }

   printf("%s: %d: Robot::setDOFacceleration(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}



//! Gets the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param x the returned value
//! \return 1 in case of success, 0 otherwise
int Robot::getDOFvalue(unsigned int index, float &x)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
//           x= joints[i]->q[j].value();
          joints[i]->getDOFvalue(j, x);
          return 1;
        }
        count++;
      }
   }


   printf("%s: %d: Robot::getDOFvalue(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}


//! Increments the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param ratio increment ratio (the increment will be ratio*(qmax-qmin))
//! \return 1 in case of success, 0 otherwise
int Robot::incrementDOFvalue(unsigned int index, float ratio)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
          joints[i]->incrementDOFvalue(j, ratio);
          updatePassiveJoints();
          updateConfiguration();
          return 1;
        }

        count++;
      }
   }

   printf("%s: %d: Robot::incrementDOFvalue(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}



//! Decrements the value of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param ratio decrement ratio (the decrement will be ratio*(qmax-qmin))
//! \return 1 in case of success, 0 otherwise
int Robot::decrementDOFvalue(unsigned int index, float ratio)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
          joints[i]->decrementDOFvalue(j, ratio);
          updatePassiveJoints();
          updateConfiguration();
          return 1;
        }

        count++;
      }
   }

   printf("%s: %d: Robot::decrementDOFvalue(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}

//! Gets the unit of one of the robot DOFs.
//! \param index index of the DOF (starts from zero and skips the DOFs in BLOCKED or CONSTRAINED state)
//! \param unit the returned unit
//! \return 1 in case of success, 0 otherwise
int Robot::getDOFunit(unsigned int index, DOFunit &unit)
{
   unsigned int i, j, count;

   //parse all the active DOFs:
   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        if(count==index) 
        {
          unit= joints[i]->q[j].unit();
          return 1;
        }
        count++;
      }
   }


   printf("%s: %d: Robot::getDOFunit(): DOF index exceeds the robot number of DOFs.\n", __FILE__, __LINE__);
   return 0;
}


//! Gets the number of DOFs of the robot (the DOFs in BLOCKED or CONSTRAINED state are not counted).
//! \return the number of DOFs
unsigned int Robot::nbDOFs()
{
   unsigned int i, j, count;

   count= 0;
   for(i=0; i<joints.size(); ++i)
   {
      for(j=0; j<joints[i]->q.size(); j++)
      {
        if(joints[i]->q[j].state==BLOCKED)
        {  continue;  }

        if(joints[i]->q[j].state==CONSTRAINED)
        {  continue;  }

        count++;
      }
   }

   return count;
}


//! Sets all the passive joints values from the current active joints values.
void Robot::updatePassiveJoints()
{
   unsigned int i;
   float a;

   for(i=0; i<constraints.size(); ++i)
   {
     constraints[i].master->getDOFvalue(constraints[i].master_dof, a);
     constraints[i].slave->setDOFvalue(constraints[i].slave_dof, a*constraints[i].parameters[0] + constraints[i].parameters[1]);
   }

}

//! Prints the current configuration of the robot.
//! The passive or constrained DOFs are skipped.
//! \param angle_unit the angles will be printed with the chosen unit
//! \return 1 in case of success, 0 otherwise
int Robot::printConfiguration(DOFunit angle_unit)
{
  unsigned int i;
  std::vector<DOF> q;

  getDOFs(q);

  printf("q= [ ");
  for(i=0; i<q.size(); ++i)
  {
    if(q[i].state==CONSTRAINED || q[i].state==BLOCKED)
    { continue; }

    switch(q[i].unit())
    {
      case METER:
         printf("%2.3f ", q[i].value());
      break;
      case RADIAN:
         if(angle_unit==RADIAN)
         {
            printf("%2.3f ", q[i].value());
            break;
         }
         if(angle_unit==DEGREE)
         {
           printf("%2.3f ", RADTODEG*q[i].value());
           break;
         }
      break; 
      case DEGREE:
         if(angle_unit==RADIAN)
         {
           printf("%2.3f ", DEGTORAD*q[i].value());
           break;
         }
         if(angle_unit==DEGREE)
         {
           printf("%2.3f ", q[i].value());
           break;
        }
     break;
    }
  }
  printf("]\n");

  return 1;
}

//! Updates the configurations of all the robot bodies from the current configuration.
void Robot::updateConfiguration()
{
  unsigned int i;
  HTMatrix pose;
  Joint *joint;
  std::list<Joint*> open;

  updatePassiveJoints();

  for(i=0; i<joints.size(); ++i)
  {
    joints[i]->computeTransform();
    joints[i]->up_to_date= false;
    joints[i]->frame.setIdentity();

    if(joints[i]->previous==NULL)
    {
      joints[i]->frame= joints[i]->Tconst * joints[i]->T;
      joints[i]->up_to_date= true;
    }
  }

  for(i=0; i<joints.size(); ++i)
  {  
     if(joints[i]->previous==NULL)
     {  open.push_back(joints[i]);  }
  }

  while(!open.empty())
  {
    joint= open.back();
    open.pop_back();
    for(i=0; i<joint->next.size(); ++i)
    {
      joint->next[i]->frame= joint->next[i]->previous->frame * joint->next[i]->Tconst * joint->next[i]->T;
      joint->next[i]->up_to_date= true;
      open.push_back(joint->next[i]);
    } 
  }

  for(i=0; i<bodies.size(); ++i)
  {
    pose= bodies[i]->joint()->frame * bodies[i]->Tjoint();
    bodies[i]->setPose(pose);
  }
}


//! Adds a joint to the robot.
//! \param name name of the joint to add
//! \param type type of the joint (see joint.h)
//! \param qmin vector of the minimal bounds on the DOF values of the joint (angles must be in degrees)
//! \param qmax vector of the maximal bounds on the DOF values of the joint (angles must be in degrees)
//! \param Tconst (constant) transformation between the previous joint frame and the new one
//! \param previous name of the previous joint in the kinematic tree
//! \return 1 in case of success, 0 otherwise
int Robot::addJoint(std::string name, JointType type, std::vector<float> qmin, std::vector<float> qmax, HTMatrix Tconst, std::string previous)
{
  bool goodSizes;
  unsigned int i;
  Joint *previous_joint= NULL, *new_joint= NULL;

  for(i=0; i<joints.size(); ++i)
  {
    if(joints[i]->name()==name)
    {
      printf("%s: %d: Robot::addJoint(): there is already a joint named \"%s\".\n", __FILE__, __LINE__, name.c_str());
      name+= "_bis";
      printf("-> the new \"%s\" is renamed \"%s\".\n", joints[i]->name().c_str(), name.c_str());
      break;
    }
  }

  if(!joints.empty())
  {
    for(i=0; i<joints.size(); ++i)
    { 
      if(joints[i]->name()==previous)
      {
        previous_joint= joints[i];
        break;
      }     
    }
    
    if(i==joints.size())
    {
      printf("%s: %d: Robot::addJoint(): the robot has no joint named \"%s\".\n", __FILE__, __LINE__, previous.c_str());
      return 0;
    }
  }

  goodSizes= true;
  switch(type)
  {
    case BASE:
      if( qmin.size()!=2 || qmax.size()!=2 )
      { goodSizes= false; }
      qmin.push_back(-180); // the function is supposed to receive angles in degrees
      qmax.push_back(180); 
    break;
    case FREEFLYER:
      if( qmin.size()!=3 || qmax.size()!=3 )
      { goodSizes= false; }
      else
      { 
         qmin.push_back(0);
         qmin.push_back(0);
         qmin.push_back(0);
         qmax.push_back(1);
         qmax.push_back(1);
         qmax.push_back(1);
      }
    break;
    case REVOLUTE: case S_REVOLUTE: case PRISMATIC:
      if( qmin.size()!=1 || qmax.size()!=1 )
      { goodSizes= false; }
    break;
  }
  if(goodSizes==false)
  { 
   printf("%s: %d: Robot::addJoint(): the size of qmin or qmax is not compatible with the type of joint \"%s\" (%s).\n", __FILE__, __LINE__, _name.c_str(),jointTypeToString(type).c_str());
   return 0;
  }

  new_joint= new Joint(type);

  if(previous_joint!=NULL)
  {
    new_joint->previous     = previous_joint;
    previous_joint->next.push_back(new_joint); 
  }
  else
  {
    new_joint->previous     = NULL;
  }
  

  for(i=0; i<new_joint->q.size(); ++i)
  {
    switch(new_joint->q[i].unit())
    {
      case METER:
        new_joint->q[i].setBounds(qmin[i], qmax[i]);
      break;
      case RADIAN:
        new_joint->q[i].setBounds(DEGTORAD*qmin[i], DEGTORAD*qmax[i]);
      break;
      case DEGREE:
        new_joint->q[i].setBounds(qmin[i], qmax[i]);
      break;
    }

    if(new_joint->q[i].min()==new_joint->q[i].max())
    {
      new_joint->q[i].state= BLOCKED;
      new_joint->q[i].setValue(new_joint->q[i].min());
    }
  }

  new_joint->rename(name);
  new_joint->Tconst= Tconst;

  if(!joints.empty())
  {
    new_joint->frame= previous_joint->frame * new_joint->Tconst;
  }
  else
  {
    new_joint->frame= new_joint->Tconst;
  }

  joints.push_back(new_joint);

  return 1;
}


//! Adds a body to the robot from a file.
//! \param name name of the body to add
//! \param type type of the body to add
//! \param file name of the file containing the 3D model of the body
//! \param scale scale factor to use with the 3D model file data
//! \param joint_name name of the joint the body is associated to
//! \param Tjoint (constant) transformation matrix between the joint frame and the body pose
//! \return 1 in case of success, 0 otherwise
int Robot::addBody(std::string name, BodyType type, std::string file, float scale, std::string joint_name, HTMatrix Tjoint)
{
  unsigned int i;
  Joint *joint= NULL;
  Body *new_body= NULL;
  
  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->name()==name)
    {
      printf("%s: %d: Robot::addBody(): there is already a body named \"%s\".\n", __FILE__, __LINE__, name.c_str());
      name+= "_bis";
      printf("-> the new \"%s\" is renamed \"%s\".\n", bodies[i]->name().c_str(), name.c_str());
      break;
    } 
  }

  if(file.empty())
  {
    printf("%s: %d: Robot::addBody(): the file name is empty.\n", __FILE__, __LINE__);
    return 0;
  }

  if(models_directory.empty())
  {
    models_directory= "./";
  }

  if(file[0]!='/') //if the file is not given in absolute path
  {
   if(models_directory.at(models_directory.size()-1)!='/')
   {   models_directory+= '/';   }
   file= models_directory + file;
  }


  joint= NULL;
  for(i=0; i<joints.size(); ++i)
  {
    if(joint_name==joints[i]->name())
    {
      joint= joints[i];
      break;
    }
  }
  if(i==joints.size())
  {
    printf("%s: %d: Robot::addBody(): robot \"%s\" has no joint named \"%s\".\n", __FILE__, __LINE__,_name.c_str(),joint_name.c_str());
    return 0;    
  }

  new_body= new Body(name, type, file, scale);
  new_body->setJoint(joint);
  new_body->setTjoint(Tjoint);
  bodies.push_back(new_body);

  return 1;
}


//! Adds an already created body to the robot.
//! \param body pointer to the body to add
//! \param joint_name name of the joint the body is associated to
//! \param Tjoint (constant) transformation matrix between the joint frame and the body pose
//! \return 1 in case of success, 0 otherwise
int Robot::addBody(Body *body, std::string joint_name, HTMatrix Tjoint)
{
  unsigned int i;
  std::string name;
  Joint *joint= NULL;
  
  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->name()==body->name())
    {
      printf("%s: %d: Robot::addBody(): there is already a body named \"%s\".\n", __FILE__, __LINE__, body->name().c_str());
      name= body->name() + "_bis";
      body->rename(name);
      printf("-> the new \"%s\" is renamed \"%s\".\n", bodies[i]->name().c_str(), body->name().c_str());
      break;
    } 
  }

  joint= NULL;
  for(i=0; i<joints.size(); ++i)
  {
    if(joint_name==joints[i]->name())
    {
      joint= joints[i];
      break;
    }
  }
  if(i==joints.size())
  {
    printf("%s: %d: Robot::addBody(): robot \"%s\" has no joint named \"%s\".\n", __FILE__, __LINE__,name.c_str(),joint_name.c_str());
    return 0;    
  }

  body->setJoint(joint);
  body->setTjoint(Tjoint);
  bodies.push_back(body);

  return 1;
}


//! Loads the robot description from an XML file.
//! \param filename name of the XML file to load
//! \return 1 in case of success, 0 otherwise
int Robot::loadFile(std::string path, std::string filename)
{
   unsigned int i;
   unsigned int length;
   size_t pos;

   reset();
 
   length= filename.length(); 
   pos= filename.find(".xml"); 
   if(pos==filename.npos || pos!=length-4)
   {
     printf("File \"%s\" does not have the good extension (it should be a .xml file).\n", filename.c_str());
     return 0;
   }    

   pos= path.find("/"); 
   if(pos==path.npos || pos!=0)
   {
     printf("Path \"%s\" must be absolute.\n", path.c_str());
     return 0;
   } 


   _path= path;
   _file= path + "/" + filename;

   joints.reserve(15);
   bodies.reserve(15);

   if(!parseRobotFile(_path, _file, this))
   {
     printf("File \"%s\" was not correctly parsed.\n", _file.c_str());
   }



   //Create collision models:
   for(i=0; i<bodies.size(); ++i)
   { 

     bodies[i]->createPQPmodel();
   }

   updateConfiguration();

   return 1;
}

//! Adds a constraint between two joints (a master and a slave) of the robot.
//! \param master name of the master joint
//! \param slave name of the slave joint
//! \param type type of the constraint (see joint.h)
//! \param master_dof index of the master joint DOF that will control the slave joint DOF (starts from 0)
//! \param slave_dof index of the slave joint DOF that will be controlled by the master joint DOF (starts from 0)
//! \param params parameters of the constraint (see joint.h)
//! \return 1 in case of success, 0 otherwise
int Robot::addConstraint(std::string master, std::string slave, ConstraintType type, unsigned int master_dof, unsigned int slave_dof, std::vector<float> params)
{
  bool nomaster= true, noslave= true;
  unsigned int i, master_index= 0, slave_index= 0;
  Constraint constraint;

  if(params.size()<2)
  {
    printf("%s: %d: Robot::addConstraint(): the size of constraint parameter array should be at least 2.\n", __FILE__, __LINE__);
    return 0;
  }

  for(i=0; i<joints.size(); ++i)
  {
    if(joints[i]->name()==master)
    {
      master_index= i;
      nomaster= false;
    }
    if(joints[i]->name()==slave)
    {
      slave_index= i;
      noslave= false;
    }
  }

  if(nomaster)
  {
    printf("%s: %d: Robot::addConstraint(): there is no joint named %s.\n", __FILE__, __LINE__, master.c_str());
    return 0;
  }

  if(noslave)
  {
    printf("%s: %d: Robot::addConstraint(): there is no joint named %s.\n", __FILE__, __LINE__, slave.c_str());
    return 0;
  }

  if(master_dof >= joints[master_index]->q.size())
  {
    printf("%s: %d: Robot::addConstraint(): the given DOF index exceeds the DOF array size of joint %s .\n", __FILE__, __LINE__, master.c_str());
    return 0;
  }

  if(slave_dof >= joints[slave_index]->q.size())
  {
    printf("%s: %d: Robot::addConstraint(): the given DOF index exceeds the DOF array size of joint %s .\n", __FILE__, __LINE__, slave.c_str());
    return 0;
  }

  if(joints[master_index]->q[master_dof].state==BLOCKED || joints[slave_index]->q[slave_dof].state==BLOCKED)
  {
    printf("%s: %d: Robot::addConstraint(): one of the DOFs of the constraint between joints \"%s\" and \"%s\" is in BLOCKED state.\n", __FILE__, __LINE__, master.c_str(), slave.c_str());
    return 0;
  }

  constraint.master= joints[master_index];
  constraint.slave = joints[slave_index];

  constraint.master_dof= master_dof;
  constraint.slave_dof = slave_dof;

  constraint.parameters.clear();
  constraint.parameters.push_back(params[0]);
  constraint.parameters.push_back(params[1]);

  joints[slave_index]->q[slave_dof].state= CONSTRAINED;


  constraints.push_back(constraint);

  return 1;
}


//! Performs a collision test for all the robot's pairs of bodies that are not linked by a joint.
//! \return true in case of collision, false otherwise
bool Robot::selfCollisionTest()
{
  unsigned int i, j;
  std::list< std::pair<Body*,Body*> >::iterator ipair; 
 
  for(i=0; i<bodies.size(); ++i)
  {
    for(j=i+1; j<bodies.size(); j++)
    {
       if(bodies[i]->joint()==bodies[j]->joint())
       {  continue;  }

       if(bodies[i]->joint()->previous==bodies[j]->joint() || bodies[j]->joint()->previous==bodies[i]->joint())
       {  continue;  }

       for(ipair=deactivatedPairs.begin(); ipair!=deactivatedPairs.end(); ++ipair)
       {
         if( (ipair->first==bodies[i] && ipair->second==bodies[j]) || (ipair->first==bodies[j] && ipair->second==bodies[i]) )
         {  break;  } 
       }
       if(ipair!=deactivatedPairs.end())
       {  continue;  }

       if(collisionTest(bodies[i], bodies[j]))
       {  
         bodies[i]->colliding= true;
         bodies[j]->colliding= true;
         //printf("collision between %s and %s.\n", bodies[i]->name().c_str(), bodies[j]->name().c_str());
         return true;
       }
    }
  }

  return false;
}


//! Computes the minimal distance between all the robot's pairs of bodies that are not linked by a joint.
//! NB: this can be computationnally expensive (and not very useful) because
//! bodies of a robot usually move very closely to each other.
//! \return the minimal distance if there is no collision, 0 otherwise
float Robot::selfCollisionDistance(Vector3D &closest_point1, Vector3D &closest_point2)
{
  bool firstTest= true;
  unsigned int i, j;
  float d, dmin= -1;
  float p1[3], p2[3];

  for(i=0; i<bodies.size(); ++i)
  {
    for(j=i+1; j<bodies.size(); j++)
    {
       if(bodies[i]->joint()==bodies[j]->joint())
       {  continue;  }

       if(bodies[i]->joint()->previous==bodies[j]->joint() || bodies[j]->joint()->previous==bodies[i]->joint())
       {  continue;  }

       d= distanceComputation(bodies[i], bodies[j], p1, p2);
       
       if(d<=0)
       {  return 0;  }

       if(d < dmin || firstTest==true)
       {
         firstTest= false;
         dmin= d;
         closest_point1.set(p1[0], p1[1], p1[2]);
         closest_point2.set(p2[0], p2[1], p2[2]);
       }
    }
  }

  return dmin;
}


//! Performs a collision test for all the robot's pairs of bodies that are not linked by a joint, when
//! the robot is moving (with a constant speed).
//! \param q robot's starting configuration (angles must be given in rad/s)
//! \param speed robot's speed (angle speeds must be given in rad/s)
//! \param time time length of the motion
//! \return true in case of collision, false otherwise
bool Robot::motionSelfCollisionTest(std::vector<float> q, std::vector<float> speed, float time)
{
  unsigned int i, j, nbSteps;
  float angle, maxAngle, translation, maxTranslation; 
  std::vector<DOF> dofs;

  if(q.size()!=speed.size())
  {
    printf("%s: %d: Robot::motionSelfCollisionTest(): vectors q and speed should have the same size.\n", __FILE__, __LINE__);
    return true;
  }

  if(q.size()!=nbDOFs())
  {
    printf("%s: %d: Robot::motionSelfCollisionTest(): vectors q and speed do not have the good size (%ud instead of %ud).\n", __FILE__, __LINE__,q.size(), nbDOFs());
    return true;
  }

  getDOFs(dofs);

  // find the maximal variations in angle and translation:
  maxAngle= 0;
  maxTranslation= 0;
  for(i=0; i<speed.size(); ++i)
  {
    switch(dofs[i].unit())
    {
      case METER:
        translation= speed[i]*time;
        if(translation > maxTranslation)
        {
           maxTranslation= translation;
        }
      break;
      case RADIAN: case DEGREE:
        angle= speed[i]*time;
        if(angle > maxAngle)
        {
           maxAngle= angle;
        }
      break;
    }
  }  

  nbSteps=  (unsigned int) ( max( fabs(maxAngle/ANGLE_STEP), fabs(maxTranslation/TRANSLATION_STEP) ) ); 

  for(i=0; i<nbSteps; ++i)
  {
    for(j=0; j<q.size(); j++)
    {
      switch(dofs[j].unit())
      {
        case METER:
          q[j]+= sign(speed[j])*TRANSLATION_STEP;
        break;
          case RADIAN: case DEGREE:
          q[j]+= sign(speed[j])*ANGLE_STEP;
        break;
      }
    }    

    setConfiguration(q);
    updateConfiguration();
    if(selfCollisionTest())
    {
      return true;
    }
  }


  return false;
}

//! Performs a collision test between all the robot's bodies and an obstacle.
//! \param obstacle the body to test the robot collision with
//! \return true in case of collision, false otherwise
bool Robot::bodyCollisionTest(Body *obstacle)
{
  unsigned int i;
 
  for(i=0; i<bodies.size(); ++i)
  {
    if(collisionTest(bodies[i], obstacle))
    {  
      bodies[i]->colliding= true;
      obstacle->colliding= true;
      //printf("collision between %s and %s.\n", bodies[i]->name().c_str(), obstacle->name().c_str());
      return true;
    }
  }

  return false;
}

//! Resets all the robot data.
//! \return 1 in case of success, 0 otherwise
int Robot::reset()
{
  _name= "none";
  bodies.clear();
  joints.clear();
  constraints.clear();
  deactivatedPairs.clear();
/*
  for(unsigned int i= 0; i<collisionPairs.size(); ++i)
  { collisionPairs[i].clear();  }
  collisionPairs.clear();
*/

  return 1;
}

//! Deactivates a collision pair.
//! \return 1 in case of success, 0 otherwise
int Robot::deactivateCollisionPair(Body *body1, Body* body2)
{
  if(body1==NULL || body2==NULL)
  {
    printf("%s: %d:  Robot::deactivateCollisionPair(): one of the input bodies is NULL.\n", __FILE__, __LINE__);
    return 0;
  }

  std::list< std::pair<Body*,Body*> >::iterator ipair;  

  for(ipair= deactivatedPairs.begin(); ipair!=deactivatedPairs.end(); ++ipair)
  {  
    if( (ipair->first==body1 && ipair->second==body2) || (ipair->first==body2 && ipair->second==body1)  )
    {
      return 0;
    }
  }
  deactivatedPairs.push_back( std::make_pair(body1, body2) );

  return 1;
}

World::World()
{
  _file= "none";
  models_directory= "./";
  robot_directory= "./";
  xmin= -2.5;
  xmax=  2.5;
  ymin= -2.5;
  ymax=  2.5;
  zmin=  0.0;
  zmax=  4.0;
  lightPosition.set(0.5*(xmin+xmax), 0.5*(ymin+ymax), zmax);
}

World::~World()
{
  unsigned int i;

  for(i=0; i<obstacles.size(); ++i)
  { 
    delete obstacles[i];
  }
  obstacles.clear();
}


//! Adds an obstacle (Body) to the world.
//! \param name name of the obstacle to add
//! \param type type of the obstacle to add
//! \param file name of the file containing the 3D model of the body
//! \param scale scale factor to use with the 3D model file data
//! \param pose transformation matrix representing the pose of the obstacle in the world frame
//! \return 1 in case of success, 0 otherwise
int World::addObstacle(std::string name, BodyType type, std::string file, float scale, HTMatrix pose)
{
  unsigned int i;
  Body *body= NULL;

  for(i=0; i<obstacles.size(); ++i)
  {
    if(obstacles[i]->name()==name)
    {
      printf("%s: %d:  World::addObstacle(): there is already an obstacle named %s.\n", __FILE__, __LINE__, name.c_str());
      name+= "_bis";
      printf("-> the new \"%s\" is renamed \"%s\".\n", obstacles[i]->name().c_str(), name.c_str());
      break;
    } 
  }


  if(file.empty())
  {
    printf("%s: %d: World::addObstacle(): the file name is empty.\n", __FILE__, __LINE__);
    return 0;
  }

  if(models_directory.empty())
  {
    models_directory= "./";
  }

  if(file[0]!='/') //if the file is not given in absolute path
  {
   if(models_directory.at(models_directory.size()-1)!='/')
   {   models_directory+= '/';   }
   file= models_directory + file;
  }

  body= new Body(name, type, file, scale);
  body->setPose(pose);
  obstacles.push_back(body);

  return 1;
}



//! Adds an already created obstacle (Body) to the world.
//! \param body pointer to the obstacle to add
//! \param pose transformation matrix representing the pose of the obstacle in the world frame
//! \return 1 in case of success, 0 otherwise
int World::addObstacle(Body *body, HTMatrix pose)
{
  unsigned int i, n;
  std::string name;

  //first find a valid (i.e.unique) name:
  name= body->name();
  n= 1;
  while(true)
  {
    for(i=0; i<obstacles.size(); ++i)
    {
      if(obstacles[i]->name()==name)
      {
        name= body->name() + "(" + convertToString(n) + ")";
        n++;
        break;
      } 
    }
    if( i >= obstacles.size() )
    {  break;  }
  }

  if(name!=body->name())
  {
    printf("%s: %d:  World::addObstacle(): there is already an obstacle named \"%s\".\n", __FILE__, __LINE__, body->name().c_str());
    body->rename(name);
    printf("-> the new obstacle is renamed \"%s\".\n", body->name().c_str());
  }


  body->setPose(pose);
  obstacles.push_back(body);

  return 1;
}


//! Removes the obstacles whose name begins like the specified string.
//! \param prefix the obstacles with this prefix will be removed
//! \return 0 in case of success, 1 otherwise
int World::removeObstacles(const std::string &prefix)
{
  if(this==NULL)
  {
    printf("%s: %d: World::removeObstacles(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return 1;
  }

  std::vector<Body*>::iterator iter;

  iter=obstacles.begin();
  while(iter!=obstacles.end())
  {
    if((*iter)->name().compare(0, prefix.length(), prefix)==0)
    { 
      delete (*iter);
      iter= obstacles.erase(iter);
      continue;
    }
    iter++;
  }

  return 0;
}

//! Draws all the world elements.
//! \param smooth use smooth (true) or flat (false) shading
//! \return 1 in case of success, 0 otherwise
int World::draw(bool smooth)
{
  unsigned int i;

  robot.draw(smooth);
 
  for(i=0; i<obstacles.size(); ++i)
  {
    obstacles[i]->draw(smooth);
  }

  return 1;
}


//! Draws the world objects with black, unlit bodies.
//! This function is meant to be used to display planar shadows.
//! \return 1 in case of success, 0 otherwise
int World::drawInBlack()
{
  unsigned int i;

  robot.drawInBlack();
 
  for(i=0; i<obstacles.size(); ++i)
  {
    obstacles[i]->drawInBlack();
  }

  return 1;
}

//! Resets all the world data.
//! \return 1 in case of success, 0 otherwise
int World::reset()
{
  robot.reset();
  obstacles.clear();

  return 1;
}

//! Loads the world description from an XML file.
//! \param filename name of the XML file to load
//! \return 1 in case of success, 0 otherwise
int World::loadFile(std::string path, std::string filename)
{
  unsigned int i, length;
  size_t pos;
 
  reset();
  length= filename.length(); 
  pos= filename.find(".xml"); 
  if(pos==filename.npos || pos!=length-4)
  {
    printf("File %s does not have the good extension (it should be a .xml file).\n", filename.c_str());
    return 0;
  }      

   pos= path.find("/"); 
   if(pos==path.npos || pos!=0)
   {
     printf("Path \"%s\" must be absolute.\n", path.c_str());
     return 0;
   } 

  _path = path;
  _file= path + "/" + filename;



  parseWorldFile(_path, _file, this);


  // Create collision models:
  for(i=0; i<obstacles.size(); ++i)
  { 
    obstacles[i]->createPQPmodel();
  }

  return 1;
}


//! Performs a self-collision test of the robot and collision tests between each obstacle and the robot.
//! \return true in case of collision, false otherwise
bool World::collisionTest()
{
  unsigned int i;

  if(robot.selfCollisionTest())
  {
    return true;
  }

  for(i=0; i<obstacles.size(); ++i)
  {
    if(robot.bodyCollisionTest(obstacles[i]))
    {   return true;   }
  }

  return false;
}


//! Performs a collision test for the world robot (selfcollision plus collision against environment obstacles),
//! for a given configuration and the configuration that will be reached after a given time period
//! for the given acceleration and velocity..
//! \param q robot's starting configuration (angles must be given in rads)
//! \param vq robot's velocity (angle velocities must be given in rad/s)
//! \param vq robot's acceleration (angle accelerations must be given in rad/s*s)
//! \param time_period time period (in seconds)
//! \return true in case of collision, false otherwise
bool World::earlyCollisionTest(float time)
{
  std::vector<float> q0, vq0, aq0, qnext;
  unsigned int i=0;
  bool result;

  robot.getConfiguration(q0);
  robot.getVelocity(vq0);
  robot.getAcceleration(aq0);
  qnext.resize(q0.size());

  for(i=0; i<q0.size(); ++i){
    qnext[i]= 0.5*time*time*aq0[i]+ time*vq0[i] + q0[i];
    //printf("q %f vq %f aq %f time %f qnext %f\n",q0[i],vq0[i], aq0[i], time, qnext[i]);  
  }
 
  result= robot.setConfiguration(qnext);
  robot.updateConfiguration();
  //robot.draw();
  if( result==0 || collisionTest())
  {  
    result= robot.setConfiguration(q0);
    robot.updateConfiguration();
    return true;
  }
  result= robot.setConfiguration(q0);
  robot.updateConfiguration();
  return false;
}

//! Computes the minimal distance between any two bodies of the world (pairs of robot bodies that
//! are not linked by a joint and pairs composed by a robot body and an environment obstacle).
//! This function gives an idea of how far the configuration is from a collision.
//! closestPoint1 closest point on the first body
//! closestPoint2 closest point on the second body
//! \return distance (>0) between the bodies if they are not colliding, -1 otherwise
float World::closestCollision(Vector3D &closestPoint1, Vector3D &closestPoint2)
{
  unsigned int i, j;
  float d, dmin= 1e9;
  float p1[3], p2[3];

  for(i=0; i<robot.bodies.size(); ++i)
  {
    robot.bodies[i]->colliding= false;
  }
  for(j=0; j<obstacles.size(); j++)
  {
    obstacles[j]->colliding= false;
  }

  for(i=0; i<robot.bodies.size(); ++i)
  {
    for(j=0; j<obstacles.size(); j++)
    {
      d= distanceComputation(robot.bodies[i], obstacles[j], p1, p2);
      if(d<0)
      {
        robot.bodies[i]->colliding= true;
        obstacles[j]->colliding= true;
        return -1;
      }
      if(d<dmin)
      {
        dmin= d;
        closestPoint1.x= p1[0];
        closestPoint1.y= p1[1];
        closestPoint1.z= p1[2];

        closestPoint2.x= p2[0];
        closestPoint2.y= p2[1];
        closestPoint2.z= p2[2];
      }
    }
  }

  for(i=0; i<robot.bodies.size(); ++i)
  {
    for(j=i+1; j<robot.bodies.size(); j++)
    {
      if(robot.bodies[i]->joint()==robot.bodies[j]->joint())
      {  continue;  }

      if(robot.bodies[i]->joint()->previous==robot.bodies[j]->joint() || robot.bodies[j]->joint()->previous==robot.bodies[i]->joint())
      {  continue;  }

      d= distanceComputation(robot.bodies[i], robot.bodies[j], p1, p2);
      if(d<0)
      {
        robot.bodies[i]->colliding= true;
        robot.bodies[j]->colliding= true;
        return -1;
      }
      if(d<dmin)
      {
        dmin= d;
        closestPoint1.x= p1[0];
        closestPoint1.y= p1[1];
        closestPoint1.z= p1[2];

        closestPoint2.x= p2[0];
        closestPoint2.y= p2[1];
        closestPoint2.z= p2[2];
      }
    }
  }


  return dmin;
}


//! Enables/disables the obstacle texture display.
//! \param on enables/disables all the obstacle textures if true/false
//! \return 1 in case of success, 0 otherwise
int World::displayObstacleTextures(bool on)
{
  unsigned int i;

  for(i=0; i<obstacles.size(); ++i)
  {
     obstacles[i]->enable_texture= on;
  }  

  return 1;
}

//! Returns a pointer to the robot's body with the specified name.
//! \param name name of the body
//! \return pointer to the body in case of success, NULL otherwise
Body* Robot::getBodyByName(std::string name)
{
  unsigned int i;

  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->name()==name)
    {
      return bodies[i];
    }
  }

  //printf("%s: %d: Robot::getBodyByName: robot \"%s\" has no body named \"%s\".\n", __FILE__,__LINE__,_name.c_str(), name.c_str());

  return NULL;
}

//! Returns a pointer to the robot's joint with the specified name.
//! \param name name of the joint
//! \return pointer to the joint in case of success, NULL otherwise
Joint* Robot::getJointByName(std::string name)
{
  unsigned int i;

  for(i=0; i<joints.size(); ++i)
  {
    if(joints[i]->name()==name)
    {
      return joints[i];
    }
  }

  //printf("%s: %d: Robot::getJointByName: robot \"%s\" has no joint named \"%s\".\n", __FILE__,__LINE__,_name.c_str(), name.c_str());

  return NULL;
}


//! Connects a robot (called sub-robot) to a joint of the calling robot.
//! The sub-robot's root joint is connected to the specified robot's joint and the desired
//! transformation between these two joints is given as a parameter.
//! \param connection name of the robot's joint the subrobot will be connected to 
//! \param T constant transformation between the "connection" joint and the sub-robot kinematic tree root
//! \param subrobot pointer to the sub-robot
//! \return 1 in case of success, 0 otherwise
int Robot::connectRobot(std:: string connection, HTMatrix &T, Robot *subrobot)
{
  if(subrobot==NULL)
  {
    printf("%s: %d: Robot::connectRobot: input subrobot is NULL.\n", __FILE__,__LINE__);
    return 0;
  }

  unsigned int i, j;
  std::string previous, name;
  std::vector<float> qmin, qmax;
  HTMatrix T2;
  Body *new_body= NULL;
  Joint *joint= NULL, *subroot= NULL;

  joint= getJointByName(connection);


  if(joint==NULL)
  {
    printf("%s: %d: Robot::connectRobot: robot \"%s\" has no joint named \"%s\".\n", __FILE__,__LINE__,_name.c_str(), connection.c_str());
    return 0;
  }

  for(i=0; i<joints.size(); ++i)
  {
    for(j=0; j<subrobot->joints.size(); j++)
    {
       if(subrobot->joints[j]->previous==NULL)
       {
         continue;
       }
       if(joints[i]->name()==subrobot->joints[j]->name())
       {
         //printf("%s: %d: Robot::connectRobot: robot \"%s\" already has a joint named \"%s\".\n", __FILE__,__LINE__,_name.c_str(), subrobot->joints[j]->name().c_str());
         name= subrobot->joints[j]->name() + "_bis";
         subrobot->joints[j]->rename(name);
         //printf("-> it is renamed \"%s\".\n", subrobot->joints[j]->name().c_str());
       }
    }
  }

  for(i=0; i<subrobot->joints.size(); ++i)
  {
    if(subrobot->joints[i]->previous==NULL)
    {
      subroot= subrobot->joints[i];
      break;
    }
  }

  for(i=0; i<bodies.size(); ++i)
  {
    for(j=0; j<subrobot->bodies.size(); j++)
    {
       if(bodies[i]->name()==subrobot->bodies[j]->name())
       {
         //printf("%s: %d: Robot::connectRobot: robot \"%s\" already has a body named \"%s\".\n", __FILE__,__LINE__,_name.c_str(), subrobot->bodies[j]->name().c_str());
         name= subrobot->bodies[j]->name() + "_bis";
         subrobot->bodies[j]->rename(name);
         //printf("-> it is renamed \"%s\".\n", subrobot->bodies[j]->name().c_str());
       }
    }
  }



  for(i=0; i<subrobot->joints.size(); ++i) 
  {
    if(subrobot->joints[i]->previous==NULL) //skip sub-robot's root joint:
    {
      continue;
    }

    if(subrobot->joints[i]->previous==subroot)
    {
      T2= T*subrobot->joints[i]->Tconst;
      previous= connection;
    }
    else
    {
      T2= subrobot->joints[i]->Tconst;
      previous= subrobot->joints[i]->previous->name();
    }

    qmin.resize(subrobot->joints[i]->q.size());
    qmax.resize(subrobot->joints[i]->q.size());
    for(j=0; j<subrobot->joints[i]->q.size(); j++)
    {
      if(subrobot->joints[i]->q[j].unit()==RADIAN)
      { 
        qmin[j]= RADTODEG*subrobot->joints[i]->q[j].min();
        qmax[j]= RADTODEG*subrobot->joints[i]->q[j].max();
      }
      else
      { 
        qmin[j]= subrobot->joints[i]->q[j].min(); 
        qmax[j]= subrobot->joints[i]->q[j].max();
      }
    }
    addJoint(subrobot->joints[i]->name(), subrobot->joints[i]->type, qmin, qmax, T2, previous);
  }


  for(i=0; i<subrobot->bodies.size(); ++i)
  {
     new_body= new Body();
     *new_body= *subrobot->bodies[i];
     if(subrobot->bodies[i]->joint()->previous==NULL)
     {
       addBody(new_body, connection, T*subrobot->bodies[i]->Tjoint());
     }
     else
     {
       addBody(new_body, subrobot->bodies[i]->joint()->name(), subrobot->bodies[i]->Tjoint());
     }
  }

  for(i=0; i<subrobot->constraints.size(); ++i)
  {
    addConstraint(subrobot->constraints[i].master->name(), subrobot->constraints[i].slave->name(), subrobot->constraints[i].type, subrobot->constraints[i].master_dof, subrobot->constraints[i].slave_dof, subrobot->constraints[i].parameters);
  }

  return 1;
}


int Robot::exportBodyModelsInJointFrames()
{
  unsigned int i;
  HTMatrix T, T2, T3, R;

  for(i=0; i<bodies.size(); ++i)
  {
    T2= inverse(bodies[i]->joint()->frame);
    T3= (bodies[i]->pose());

    T= T2*T3;
    bodies[i]->mesh().vertexTransform(T);
   
    //invert X and Z axis:
//     R.setRotationFromAngleAndAxis(M_PI_2, 0.0, 1.0, 0.0);
//     bodies[i]->mesh.vertexTransform(R);

    bodies[i]->mesh().saveAsOBJfile(bodies[i]->name() + "2.obj", 1000.0);
  }

  return 1;
}

#if defined(ENABLE_DISPLAY)
//! Exports the world scene to a .pov file (for POVRAY ray tracer).
//! \param camera the camera that will be exported
//! \param foldername the name of the folder where the file will be created
//! \param filename name of the file (.pov) to create
//! \param highQuality selects the quality settings of the POVRay rendering (high or low) 
//! \return 1 in case of success, 0 otherwise
int World::exportSceneToPOVRay(Camera &camera, std::string foldername, std::string filename, bool highQuality)
{
  unsigned int i;;
  size_t pos;
  float width, height, aspect;
  GLint viewport[4];
  Vector3D lookat;
  std::string script_name, inc_name;
  FILE *file= NULL;
  Body *body= NULL;
  
  if(filename.compare(filename.length()-4, 4, ".pov") != 0)
  {
    filename+= ".pov";
  }

  file= fopen((foldername+filename).c_str(), "w");
  if(file==NULL)
  {
    printf("%s: %d: World::exportSceneToPOVRay: can not open file \"%s\".\n",__FILE__,__LINE__, filename.c_str()); 
    return 0;
  }

  fprintf(file, "#include \"colors.inc\"\n");
  fprintf(file, "#include \"textures.inc\"\n");
  fprintf(file, "#include \"metals.inc\"\n");
  fprintf(file, "#include \"glass.inc\"\n");


  // Declare materials:
  fprintf(file, "#declare M_PLASTIC=\n");
  fprintf(file, " material {\n");
  fprintf(file, "\t texture {\n");
  fprintf(file, "\t\t pigment { rgbt 1 }\n");
  fprintf(file, "\t\t\t finish {\n");
  fprintf(file, "\t\t\t\t ambient 0\n");
  fprintf(file, "\t\t\t\t diffuse 0.65\n");
  fprintf(file, "\t\t\t\t roughness 0.05\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t }\n");
  fprintf(file, " }\n\n");

  fprintf(file, "#declare M_SHINY_PLASTIC=\n");
  fprintf(file, " material {\n");
  fprintf(file, "\t texture {\n");
  fprintf(file, "\t\t pigment { rgbt 1 }\n");
  fprintf(file, "\t\t\t finish {\n");
  fprintf(file, "\t\t\t\t ambient 0.1\n");
  fprintf(file, "\t\t\t\t diffuse 0.8\n");
  fprintf(file, "\t\t\t\t specular 0.8\n");
  fprintf(file, "\t\t\t\t phong 1\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t }\n");
  fprintf(file, " }\n\n");


  fprintf(file, "#declare M_GLASS=\n");
  fprintf(file, " material {\n");
  fprintf(file, "\t texture {\n");
  fprintf(file, "\t\t pigment { rgbt 1 }\n");
  fprintf(file, "\t\t\t finish {\n");
  fprintf(file, "\t\t\t\t ambient 0.0\n");
  fprintf(file, "\t\t\t\t diffuse 0.05\n");
  fprintf(file, "\t\t\t\t specular 0.6\n");
  fprintf(file, "\t\t\t\t roughness 0.005\n");
  fprintf(file, "\t\t\t\t reflection { 0.1, 1.0 fresnel on }\n");
  fprintf(file, "\t\t\t\t conserve_energy\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t }\n");
  fprintf(file, "\t interior {\n");
  fprintf(file, "\t\t ior 1.5\n");
  fprintf(file, "\t\t fade_power 1001\n");
  fprintf(file, "\t\t fade_distance 0.9\n");
  fprintf(file, "\t\t fade_color <0.7,0.7,0.7>\n");
  fprintf(file, "\t }\n");
  fprintf(file, " }\n\n");


  fprintf(file, "#declare M_METAL=\n");
  fprintf(file, " material {\n");
  fprintf(file, "\t texture {\n");
  fprintf(file, "\t\t pigment { rgbt 1 }\n");
  fprintf(file, "\t\t\t finish {\n");
  fprintf(file, "\t\t\t\t ambient 0.0\n");
  fprintf(file, "\t\t\t\t diffuse 0.3\n");
  fprintf(file, "\t\t\t\t specular 0.6\n");
  fprintf(file, "\t\t\t\t roughness 0.005\n");
  fprintf(file, "\t\t\t\t reflection { 0.2 metallic }\n");
  fprintf(file, "\t\t\t\t conserve_energy\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t }\n");
  fprintf(file, " }\n\n");

  fprintf(file, "#declare M_MARBLE=\n");
  fprintf(file, " material {\n");
  fprintf(file, "\t texture {\n");
  fprintf(file, "\t\t pigment {\n");
  fprintf(file, "\t\t\t marble\n");
  fprintf(file, "\t\t\t turbulence 0.5 omega 0.7 rotate -40*y scale 6\n");
  fprintf(file, "\t\t\t color_map {\n");
  fprintf(file, "\t\t\t\t [0.50 color rgb 1.0]\n");
  fprintf(file, "\t\t\t\t [0.57 color rgb 0.8]\n");
  fprintf(file, "\t\t\t\t [0.60 color rgb <0.9,0.8,0.7>]\n");
  fprintf(file, "\t\t\t\t [0.63 color rgb 1.0]\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t\t }\n");
  fprintf(file, "\t\t\t finish {\n");
  fprintf(file, "\t\t\t\t ambient 0.0\n");
  fprintf(file, "\t\t\t\t reflection 0.1\n");
  fprintf(file, "\t\t\t }\n");
  fprintf(file, "\t }\n");
  fprintf(file, " }\n\n");

  // Create the .inc files for each obstacle:
  for(i=0; i<obstacles.size(); ++i)
  {
    if(!obstacles[i]->enable_display)
    {  continue;  }
    obstacles[i]->exportMeshToPOVRay(foldername);
  }
  for(i=0; i<robot.bodies.size(); ++i)
  {
    if(!robot.bodies[i]->enable_display)
    {  continue;  }
    robot.bodies[i]->exportMeshToPOVRay(foldername);
  }

  //global settings:
  fprintf(file, "global_settings {\n");
  fprintf(file, "\t assumed_gamma 1\n");
  fprintf(file, "\t radiosity {\n");
  fprintf(file, "\t\t pretrace_start 0.08\n");
  fprintf(file, "\t\t pretrace_end  0.04\n");
  fprintf(file, "\t\t count 50\n");
  fprintf(file, "\t\t nearest_count 5\n");
  fprintf(file, "\t\t error_bound 0.1\n");
  fprintf(file, "\t\t recursion_limit 1\n");
  fprintf(file, "\t\t low_error_factor .5\n");
  fprintf(file, "\t\t gray_threshold 0.0\n");
  fprintf(file, "\t\t minimum_reuse 0.015\n");
  fprintf(file, "\t\t brightness 0.8\n");
  fprintf(file, "\t\t normal on\n");
  fprintf(file, "\t\t media on\n");
  fprintf(file, "\t\t adc_bailout 0.01/2\n");
  fprintf(file, "\t }\n");
  fprintf(file, "}\n\n");
 ////////////////////////////////////////////////

  //camera settings:
  glGetIntegerv(GL_VIEWPORT, viewport);
  width=  (float) (viewport[2]);
  height= (float) (viewport[3]);
  aspect= width/height;
  lookat= camera.position + camera.viewDir;


  fprintf(file, "camera {\n"); 
  camera.updateCartesianCoordinates();
  fprintf(file, "\t location  <%f, %f, %f>\n", camera.position.x, camera.position.y, camera.position.z); 
//   fprintf(file, "\t /*right  -x*image_width/image_height*/\n"); 
  fprintf(file, "\t sky  <%f, %f, %f>\n", camera.upVector.x, camera.upVector.y, camera.upVector.z ); 
// camera.rightVector*= -1.0;
   fprintf(file, "\t right  <%f, %f, %f>\n", aspect*camera.rightVector.x, aspect*camera.rightVector.y, aspect*camera.rightVector.z ); 
  // fprintf(file, "\t direction  <%f,%f,%f>\n",camera.viewDir.x, camera.viewDir.y, camera.viewDir.z); 
// printf("%f %f %f\n", dotProduct(camera.upVector,camera.rightVector),dotProduct(camera.viewDir,camera.rightVector), dotProduct(camera.upVector,camera.viewDir));
  //fprintf(file, "\t angle  %f\n", camera.fov); 
  fprintf(file, "\t look_at  <%f, %f, %f>\n", camera.focusPoint.x, camera.focusPoint.y, camera.focusPoint.z); 
  //fprintf(file, "\t look_at  <%f %f %f>\n", lookat.x, lookat.y, lookat.z); 
  fprintf(file, "}\n\n"); 



  fprintf(file, "light_source {\n"); 
  fprintf(file, "\t <%f, %f, %f>*1000\n", lightPosition.x, lightPosition.y, lightPosition.z);
  fprintf(file, "\t color rgb <1, 1, 1>\n"); 
  fprintf(file, "}\n\n"); 

  //draw the floor:
//   fprintf(file, "box {\n"); 
//   fprintf(file, "\t <%f, %f, %f>,\n", xmin, ymin, zmin-1.0); 
//   fprintf(file, "\t <%f, %f, %f>\n", xmax, ymax, zmin); 
//   fprintf(file, "\t pigment {\n"); 
//   fprintf(file, "\t\t color rgb <0.4, 0.7, 0.7>\n"); 
//   fprintf(file, "\t }\n"); 
//   fprintf(file, "\t finish {\n");
//   fprintf(file, "\t\t ambient rgb <0.0, 0.0, 0.0>\n");
//   fprintf(file, "\t\t diffuse 0.70\n");
//   fprintf(file, "\t\t brilliance 1.00\n");
//   fprintf(file, "\t\t metallic 0.9\n");
//   fprintf(file, "\t\t specular 1.0\n");
//   fprintf(file, "\t\t roughness 0.01\n");
//   fprintf(file, "\t }\n");
//   fprintf(file, "}\n\n"); 

//   fprintf(file, "plane {\n"); 
//   fprintf(file, "\t z, -0.0\n"); 
//   fprintf(file, "\t material { M_SHINY_PLASTIC }\n");
//   fprintf(file, "}\n\n"); 

  //fprintf(file, "background { color SkyBlue }\n");
  fprintf(file, "background { color White*0.50 }\n");

  //Call the .inc files:
  for(i=0; i<obstacles.size(); ++i)
  {
    body= obstacles[i];
    if(!body->enable_display)
    {  continue;  }
    body->addToPOVfile(file);
  }   

  for(i=0; i<robot.bodies.size(); ++i)
  {
    body= robot.bodies[i];
    if(!body->enable_display)
    {  continue;  }
    body->addToPOVfile(file);
    // Replace all the '.' characters by '_' the file names:
    inc_name= body->name();
    pos= inc_name.find(".");
    while(pos!=std::string::npos)
    {
     inc_name.replace(pos, 1, "_");
     pos= inc_name.find(".");
    }
  }   


  fclose(file); 

  script_name= foldername + "render";
  file= fopen(script_name.c_str(), "w");
  if(file==NULL)
  {
    printf("%s: %d: World::exportSceneToPOVRay: can not open file \"%s\".\n",__FILE__,__LINE__,script_name.c_str()); 
    return 0;
  }
  
  if(highQuality)
  {
    fprintf(file, "#!/bin/sh\n");
    fprintf(file, "~/povray-3.6/povray +I%s", filename.c_str());
    fprintf(file, " +O./outHi.png +FN16");
    fprintf(file, " +A +W%d +H%d +Q9 -D\n", (unsigned int) (width), (unsigned int) (height));
  }
  else
  {
    width/= 2.0;
    height/= 2.0;
    fprintf(file, "#!/bin/sh\n");
    fprintf(file, "~/povray-3.6/povray +I%s", filename.c_str());
    fprintf(file, " +O./outLo.png +FN16");
    fprintf(file, " +W%d +H%d +Q2 +p\n", (unsigned int) (width), (unsigned int) (height));
  }
  
  fclose(file);

  return 1;
}
#endif

//! Enables the display of all of the robot's bodies that are used for collision tests.
//! \return 1 in case of success, 0 otherwise
int Robot::enableDisplayOfCollisionBodiesOnly()
{
  unsigned int i;

  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->type()==COLLISION_ONLY || bodies[i]->type()==DISPLAY_AND_COLLISION)
    {   bodies[i]->enable_display= true;    }
    else
    {   bodies[i]->enable_display= false;   }
  }

  return 1;
}

//! Enables/disables the display of all of the robot's bodies according to their BodyType fields.
//! \return 1 in case of success, 0 otherwise
int Robot::resetDisplayStates()
{
  unsigned int i;

  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->type()==DISPLAY_ONLY || bodies[i]->type()==DISPLAY_AND_COLLISION)
    {   bodies[i]->enable_display= true;    }
    else
    {   bodies[i]->enable_display= false;   }
  }

  return 1;

}

//! Enables/disables the display of all of the robot's bodies that were originally meant to be displayed.
//! \param on set to true/false to enable/disable display
//! \return 1 in case of success, 0 otherwise
int Robot::enableDisplay(bool on)
{
  unsigned int i;

  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->type()==DISPLAY_ONLY || bodies[i]->type()==DISPLAY_AND_COLLISION)
    {   bodies[i]->enable_display= on;    }
  }

  return 1;
}

//! Enables/disables the collision test for all of the robot's bodies that were originally meant to be tested against collision.
//! \param on set to true/false to enable/disable collision test
//! \return 1 in case of success, 0 otherwise
int Robot::enableCollision(bool on)
{
  unsigned int i;

  for(i=0; i<bodies.size(); ++i)
  {
    if(bodies[i]->type()==COLLISION_ONLY || bodies[i]->type()==DISPLAY_AND_COLLISION)
    {   bodies[i]->enable_collision= on;    }
  }

  return 1;
}


//! Enables the display of all of the bodies (robot and environment) that are used for collision tests.
//! \return 0 in case of success, 1 otherwise
int World::enableDisplayOfCollisionBodiesOnly()
{
  unsigned int i;

  robot.enableDisplayOfCollisionBodiesOnly();

  for(i=0; i<obstacles.size(); ++i)
  {
    if(obstacles[i]->type()==COLLISION_ONLY || obstacles[i]->type()==DISPLAY_AND_COLLISION)
    {   obstacles[i]->enable_display= true;    }
    else
    {   obstacles[i]->enable_display= false;   }
  }

  return 0;
}

//! Enables/disables the display of all of the robot's bodies according to their BodyType fields.
//! \return 0 in case of success, 1 otherwise
int World::resetDisplayStates()
{
  unsigned int i;

  robot.resetDisplayStates();

  for(i=0; i<obstacles.size(); ++i)
  {
    if(obstacles[i]->type()==DISPLAY_ONLY || obstacles[i]->type()==DISPLAY_AND_COLLISION)
    {   obstacles[i]->enable_display= true;    }
    else
    {   obstacles[i]->enable_display= false;   }
  }

  return 0;
}



